/* io.c - Input/Output routines of qhull application

   Author:   Hannu Huhdanpaa
   Internet: hannu@geom.umn.edu

   copyright (c) 1993, The Geometry Center.

   07/09/93 - created (hh)

   */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "set.h"
#include "poly.h"
#include "qhull.h"
#include "globals.h"
#include "io.h"
#include "geom.h"





/*-----------------------------------------
-printincidences- print the facets, each identified by which vertices it
 contains
*/
void printincidences(FILE *fp, facetT *facetlist) {
  vertexT *vertex, **vertexp;
  facetT *facet;
  int numfacets= 0;

  FORALLfacet_(facetlist) {
    if (DELAUNAY && facet->normal[hull_dim-1] > 0.0)
      continue;
    numfacets++;
  }
  
  
  fprintf(fp, "Number of facets: %d\n", numfacets);
  FORALLfacet_(facetlist) {
    if (DELAUNAY && facet->normal[hull_dim-1] > 0.0)
      continue;
    if (facet->toporient) {
      FOREACHvertex_(facet->vertices) {
	fprintf(fp, "%d ", pointid_(vertex->point));
      }
    }else {
      FOREACHvertexreverse12_(facet->vertices) {
	fprintf(fp, "%d ", pointid_(vertex->point));
      }
    }
    fprintf(fp, "\n");
  }
  
} /* printincidences */









/*-------------------------------------------
-readpoints- read points from fin into all_points
    fin is lines of coordinates, one per vertex, first line number of points
returns:
    number of points, array of point coordinates, dimension
*/
coordT *readpoints(int *numpoints, int *dimension) {
  coordT *points, *coords, paraboloid;
  char line[MAXLINE+1], token[MAXLINE+1], *tokp, chr, *linep= NULL;
  int diminput, tokcount, pointindex= 0;

  
  while((linep= fgets(line, MAXLINE, fin)) && !strcmp(line, "\n"))
    ;
  if (!linep) {
    errexit(NULL, NULL, NULL, NULL);
  }
  if (!(diminput= atoi(strtok(line,WHITESPACE)))) {
    errexit(NULL, NULL, NULL, NULL);
  }
  *dimension= DELAUNAY ? diminput+1 : diminput;
  while((linep= fgets(line, MAXLINE, fin)) && !strcmp(line, "\n"))
    ;
  if (!linep) {
    errexit(NULL, NULL, NULL, NULL);
  }
  if (!(*numpoints= atoi(strtok(line,WHITESPACE)))) {
    errexit(NULL, NULL, NULL, NULL);
  }
  if (!(coords=points=(coordT*)malloc(*numpoints**dimension*sizeof(coordT)))){
    errexit(NULL, NULL, NULL, NULL);
  }
  
  
  
  
  tokp= token;
  while (linep= fgets(line, MAXLINE, fin)) {
    if (!strcmp(linep, "\n") || (pointindex++ >= *numpoints))
      continue;
    tokcount= 0;
    paraboloid= 0.0;
    while(chr= *linep++) {
      if (isspace(chr)) {
        if (tokp != token) {
          *tokp= '\0';
          *coords= atof(token);
	  if (DELAUNAY)
	    paraboloid += (*coords) * (*coords);
	  coords++;
          tokp= token;
          tokcount++;
        }
      }else
        *tokp++= chr;
    }
    if (tokp != token) {
      *tokp= '\0';
      *coords= atof(token);
      if (DELAUNAY)
	paraboloid += (*coords) * (*coords);
      coords++;
      tokcount++;
    }
    if (DELAUNAY)
      *coords++= paraboloid;
    if (tokcount != diminput) {
      errexit(NULL, NULL, NULL, NULL);
    }
  }
  if (pointindex != *numpoints) {
    errexit(NULL, NULL, NULL, NULL);
  }





  return(points);
} /* readpoints */



