/* Modified version of qhull 1.0 By Elias Tsiantas.[24/02/2004]
   -The reason of this modification is to be able to create convex
   hulls from within an application without to call a console.
 
   -This is the mini version of the previous version of the dll.
     
   -All error output code has been removed [ since it is a dll ].
   -All statistics information has been removed.
   -All ".off" output has been removed.
   -Only face indices are outputed.
   -No need to define Unix. It compiles fine with Devc++4.9.8.7
   -main function removed
   -initialize functiom renamed to qh.
   -qhull function converted form void to int to be able to return
    error codes;
*/


/*---------------------------------------------------------------*/


/* qhull - quick hull see qhull.h
   
   This file is consists of MacIntosh specific code and UNIX specific code.
   If you compile this program for the UNIX environment, keep #define UNIX,
   otherwise, delete this line.
   
   Author  : Hannu Huhdanpaa
   Internet: hannu@geom.umn.edu
   
   copyright (c) 1993, The Geometry Center
        
   02/06/93 - initial version (hh)
        
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <ctype.h>
#include "set.h"
#include "poly.h"
#include "geom.h"
#include "qhull.h"
#include "globals.h"
#include "mem.h"
#include "io.h"


extern void *buffer_begin; /* pointer to the beginning of current buffer */ 
extern int *indextable;    /* table of freelists, indexed by size */



/*-------------------------------------------------
-qhull- general dimension convex hull
*/


int qhull(char *inputfile,char *outputfile) {
  pointT *points, *minx, *maxx;
  int k, numpoints;
  setT *maxpoints, *vertices;
  facetT *facetlist, *facet;
  void *buffer, *nextbuffer;
  
  if (inputfile) {
    fout= fopen(outputfile, "w");
     if (!(fin= fopen(inputfile, "r"))) {
      //input file could not be opened
      fclose(fout);
      return-1;
    }
  }



   
  first_point= points= readpoints(&numpoints, &hull_dim);
  if (hull_dim == 1) {return-2;}
    
 

    
  normal_size= hull_dim * sizeof(coordT);
  constructsizetable();
  facet_points= (coordT *)memalloc(hull_dim * hull_dim * sizeof(coordT), &k);
  gm_matrix= (coordT *)memalloc(hull_dim * hull_dim * sizeof(coordT), &k);
  gm_row= (coordT **)memalloc(hull_dim * sizeof(coordT *), &k);
  newhashtable(hull_dim * hull_dim * 10);
  maxpoints= maxmin(points, numpoints, hull_dim, &minx, &maxx);
  k= hull_dim + 1;
  vertices= setnew(&k);
  setprepend(&vertices, newvertex(minx));
  setprepend(&vertices, newvertex(maxx));
  initialvertices(&vertices, maxpoints, points, numpoints, hull_dim+1);
 
 
    
   if (!APPROXhull)
    DELTAvalue= DISTround;
  facetlist= initialhull(vertices);
  partitionall(facetlist, vertices, points, numpoints);
  setfree(&maxpoints);
  setfree(&vertices);
  buildhull(&facetlist, numpoints);
  

  if (num_coplanar || num_nonconvex || num_nearlysingular) {
    checkorientation(facetlist, ALGORITHMfault);
    checkplanes(facetlist, ALGORITHMfault);
    checkpolygon(facetlist);
  }else
    checkorientation(facetlist, ALGORITHMfault);
  

    printincidences(fout, facetlist);// >-----------<

     
  FORALLfacet_(facetlist)
    setfree(&(facet->outsideset));
  free(points);
  


/* Memory freeing problem [solved by freeing stuff in the main qh function]


  free(hash_table);
  
  for(buffer= buffer_begin; buffer; buffer= nextbuffer) {
   nextbuffer= *((void **)buffer);
    free(buffer); }
  
  free(indextable);
*/

  if (inputfile){fclose(fin); fclose(fout);}
  
  return 1;
  
} 






/*-------------------------------------------------
-buildhull- constructs a hull by adding points to a simplex
*/
void buildhull(facetT **facetlist, int numpoints) {
  facetT *facet, **facetp, *newfacet, *newfacets;
  setT *horizon;         /* set of horizon neighbors */
  setT *interior;        /* set of visible facets for furthest */
  pointT *furthest;
  int dummy;
  

  FORALLfacet_(*facetlist) {
    if (!(furthest= SETlast_(facet->outsideset, dummy)))
      continue;
    current_furthest= pointid_(furthest);
    first_newfacet= facet_id; 
    SETdellast_(facet->outsideset, dummy);  
    horizon= findhorizon(furthest, facet, &interior);
    newfacets= makecone(furthest, horizon, interior);
    FORALLnewfacet_(newfacets)
      setfacetplane(newfacet);
    partitioninterior(newfacets, interior);
    if (BESTfurthest)
      partitionhorizonpoints(newfacets, horizon);
    deleteinterior(facetlist, interior, horizon);
    setfree(&horizon);
    setfree(&interior);
  }

} /* buildhull */






/*-------------------------------------------
-errexit- return to system after an error
prints useful information
*/
void errexit(facetT *facet, ridgeT *ridge, vertexT *vertex, pointT *point) {
  exit(1);
} /*errexit */


/*-------------------------------------------------
-findhorizon- find the horizon and interior for a point that is above the facet
returns:
     non-empty set of horizon facets
     interior: visible facets for the point
     sets facet->interior= True, facet->horizon= False for interior facets
     sets facet->horizon= True, facet->interior= False for horizon facets
*/
setT *findhorizon(pointT *point, facetT *facet, setT **interior) {
  facetT *neighbor, **neighborp;
  setT *horizon= NULL;
  coordT dist;
  int interiorcnt= 1, horizoncnt= 0, n;
  
  *interior= NULL;
  setappend(interior, facet);
  facet->interior= True; facet->horizon= False;
  facet->visitid= ++visit_id;
  SETsize_(facet->neighbors, n);
  for(n= 0; facet= SETelem_(n, facetT, *interior); n++) {
    FOREACHneighbor_(facet) {
      if (neighbor->visitid == visit_id)
	continue;
      neighbor->visitid= visit_id;
      num_visibility++;
      if ((dist= distplane(point, neighbor)) > DISTround) {
	setappend(interior, neighbor);
	interiorcnt++;
	neighbor->interior= True; neighbor->horizon= False;
      }else {

	setappend(&horizon, neighbor);
	horizoncnt++;
	neighbor->interior= False; neighbor->horizon= True;
      }
    }
  }
  if (!horizon) {
    errexit(NULL, NULL, NULL, NULL);
  }
  SETsize_(horizon, n);
  if ((hull_dim - 1) * n > 2 * HASHTABLEsize) {
    newhashtable(3 * HASHTABLEsize);
  }

  return (horizon);
} /* findhorizon */


/*-------------------------------------------------
-initialhull- constructs the initial hull as a hull_dim simplex of vertices
  returns:
    facetlist for hull
    interior_point= arithmetic center of vertices
    */
facetT *initialhull(setT *vertices) {
  facetT *facet, *facetlist, *firstfacet;
  
  facetlist= createsimplex(vertices);
  interior_point= getcenter(vertices, hull_dim+1);
  firstfacet= facetlist;
  setfacetplane(firstfacet);
  if (distplane(interior_point, firstfacet) > 0)
    fliporient(facetlist);
  FORALLfacet_(facetlist)
    setfacetplane(facet); 
  checkorientation(facetlist, DATAfault);
  checkpolygon(facetlist);
  checkplanes(facetlist, DATAfault);

  return (facetlist);
} /* initialhull */









//---------------------------------------------


__declspec(dllexport) int _stdcall qh(char *finb,
                                      char *foutb,
                                      float delta,
                                      int pf,
                                      int fo) {



//protect against false inputs
if (pf>1){pf=1;}
if (pf<0){pf=0;}
if (fo>1){fo=1;}
if (fo<0){fo=0;}
if (delta<0.0){delta=0.0;}




//partition the horizon points also=false
BESTfurthest= False;
BESTnewfacet= False;
DELAUNAY= False;


//approximate convex hull
APPROXhull= False;

FORCEoutput= False;
PRINTincidences= False;



      //approximate convex hull, delta value n
      if (delta>0.0){
        DELTAvalue= delta;
          APPROXhull= True;};
      
      //print only facets indices
      if (pf==1){
      PRINTincidences= True;}
      
      // force output despite degenerate results     
      if (fo==1){
      FORCEoutput= True;}


  if (APPROXhull) {
    if (BESTnewfacet == False) {
      BESTfurthest= True;
      BESTnewfacet= True;}}


  qhull(finb,foutb);

  free(indextable);
  
  
 return(1);     
  
} 





  



/*-------------------------------------------------
-partitionall- partitions all points into the outsidesets for facetlist
   vertices= set of vertices used by facetlist
*/
void partitionall(facetT *facetlist, setT *vertices,
		  pointT *points, int numpoints){
  setT *vertexpoints= NULL; 
  vertexT *vertex, **vertexp;
  pointT *point, *pointend;
  int upperlimit;
  
  FOREACHvertex_(vertices)
    setadd(&vertexpoints, vertex->point);
  upperlimit= setsize(vertexpoints) - 1;
  FORALLpoint_(points, numpoints)
    if (!setin(vertexpoints, point, upperlimit))
      partitionpoint(point, facetlist);
  setfree(&vertexpoints);
} /* partitionall */


/*-------------------------------------------------
-partitionhorizon- partitions the outsideset of one horizonfacet
*/
void partitionhorizon(facetT *horizonfacet, facetT *newfacets) {
  int dummy;
  pointT *point, **pointp;
  facetT *facet, *bestfacet;
  coordT dist, bestdist, currentdist;
  boolT furthestdeleted= False;
  pointT *furthest= SETlast_(horizonfacet->outsideset, dummy);
  setT *pointstodelete= NULL;
  
  FOREACHpoint_(horizonfacet->outsideset) {
    bestfacet= horizonfacet;
    num_distpart++;
    bestdist= distplane(point, horizonfacet);
    FORALLfacet_(newfacets) {
      num_distpart++;
      if ((dist= distplane(point, facet)) > bestdist) {
	bestfacet= facet;
	bestdist= dist;
      }
    }
    if (bestfacet != horizonfacet) {
      setappend(&pointstodelete, point);
      if (point == furthest)
	furthestdeleted= True;
      if (bestfacet->furthestdist < bestdist) {
	setappend(&(bestfacet->outsideset), point);
	bestfacet->furthestdist= bestdist;
      }else
	setappend2ndlast(&(bestfacet->outsideset), point);
    }
  }
  FOREACHpoint_(pointstodelete)
    setdel(&(horizonfacet->outsideset), point);
  setfree(&pointstodelete);
  if (furthestdeleted && (furthest=SETlast_(horizonfacet->outsideset, dummy))){
    num_distpart++;
    currentdist= distplane(furthest, horizonfacet);
    FOREACHpoint_(horizonfacet->outsideset) {
      num_distpart++;
      if ((dist= distplane(point, horizonfacet)) > currentdist) {
	furthest= point;
	currentdist= dist;
      }
    }
    setdel(&(horizonfacet->outsideset), furthest);
    setappend(&(horizonfacet->outsideset), furthest);
  }

} /* partitionhorizon */


/*-------------------------------------------------
-partitionhorizonpoints- partitions the points of horizon facets if needed
*/
void partitionhorizonpoints(facetT *newfacets, setT *horizon){
  facetT *facet, **facetp;;
  
  FOREACHfacet_(horizon) {
    partitionhorizon(facet, newfacets);
  }

} /* partitionhorizonpoints */


/*-------------------------------------------------
-partitioninterior- partitions points to the outside sets of facets
*/
void partitioninterior(facetT *facetlist, setT *interior) {
  facetT *facet, **facetp;
  pointT *point, **pointp;
  
  FOREACHfacet_(interior) {
    if (facet->outsideset) {
      FOREACHpoint_(facet->outsideset) {
	partitionpoint(point, facetlist);
      }
    }
  }

} /* partitioninterior */


/*-------------------------------------------------
-partitionpoint- assigns point to a visible facet in facetlist
    BESTnewfacet= assigns point to facet it is furthest above
    otherwise, picks the first one 

    DELTAvalue == DISTround in non-approximate case
*/
void partitionpoint(pointT *point, facetT *facetlist) {
  coordT dist, bestdist= DELTAvalue;
  facetT *facet, *bestfacet= NULL;
  
  num_partition++;
  FORALLfacet_(facetlist) {
    num_distpart++;
    if ((dist= distplane(point, facet)) > bestdist) {
      bestfacet= facet;
      bestdist= dist;
      if (!BESTnewfacet && bestdist > DISTround)
	break;
    }
  }
 
   if (bestfacet) {

    if (bestfacet->furthestdist < bestdist) {
      setappend(&(bestfacet->outsideset), point);
      bestfacet->furthestdist= bestdist;
    }else {
      setappend2ndlast(&(bestfacet->outsideset), point);
    }
  }

} /* partitionpoint */



